import asyncio
import concurrent.futures
import copy
import datetime
import gzip
import typing
from collections import OrderedDict
from dataclasses import dataclass
from unittest import mock

import pytest

from ydb import issues
from . import datatypes, topic_reader_asyncio
from .datatypes import PublicBatch, PublicMessage
from .topic_reader import PublicReaderSettings
from .topic_reader_asyncio import ReaderStream, ReaderReconnector, TopicReaderError
from .._grpc.grpcwrapper.common_utils import SupportedDriverType, ServerStatus
from .._grpc.grpcwrapper.ydb_topic import (
    StreamReadMessage,
    Codec,
    OffsetsRange,
    UpdateTokenRequest,
    UpdateTokenResponse,
)
from .._topic_common.test_helpers import (
    StreamMock,
    wait_condition,
    wait_for_fast,
    WaitConditionError,
)

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from .._grpc.v4.protos import ydb_status_codes_pb2
else:
    from .._grpc.common.protos import ydb_status_codes_pb2

from .._constants import DEFAULT_INITIAL_RESPONSE_TIMEOUT


@pytest.fixture(autouse=True)
def handle_exceptions(event_loop):
    def handler(loop, context):
        print(context)

    event_loop.set_exception_handler(handler)


@pytest.fixture()
def default_executor():
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=2, thread_name_prefix="decoder_executor")
    yield executor
    executor.shutdown()


def stub_partition_session(id: int = 0, ended: bool = False):
    partition_session = datatypes.PartitionSession(
        id=id,
        state=datatypes.PartitionSession.State.Active,
        topic_path="asd",
        partition_id=1,
        committed_offset=0,
        reader_reconnector_id=415,
        reader_stream_id=513,
    )

    if ended:
        partition_session.end()

    return partition_session


def stub_message(id: int):
    return PublicMessage(
        seqno=id,
        created_at=datetime.datetime(2023, 3, 18, 14, 15),
        message_group_id="",
        session_metadata={},
        offset=0,
        written_at=datetime.datetime(2023, 3, 18, 14, 15),
        producer_id="",
        data=id,
        metadata_items={},
        _partition_session=stub_partition_session(),
        _commit_start_offset=0,
        _commit_end_offset=1,
    )


@pytest.fixture()
def default_reader_settings(default_executor):
    return PublicReaderSettings(
        consumer="test-consumer",
        topic="test-topic",
        decoder_executor=default_executor,
    )


class StreamMockForReader(StreamMock):
    async def receive(self) -> StreamReadMessage.FromServer:
        return await super(self).receive()

    def write(self, message: StreamReadMessage.FromClient):
        return super().write(message)


@pytest.mark.asyncio
class TestReaderStream:
    default_batch_size = 1
    partition_session_id = 2
    partition_session_committed_offset = 10
    second_partition_session_id = 12
    second_partition_session_offset = 50
    default_reader_reconnector_id = 4

    @pytest.fixture()
    def stream(self):
        return StreamMock()

    @pytest.fixture()
    def partition_session(
        self, default_reader_settings, stream_reader_started: ReaderStream
    ) -> datatypes.PartitionSession:
        partition_session = datatypes.PartitionSession(
            id=2,
            topic_path=default_reader_settings.topic,
            partition_id=4,
            state=datatypes.PartitionSession.State.Active,
            committed_offset=self.partition_session_committed_offset,
            reader_reconnector_id=self.default_reader_reconnector_id,
            reader_stream_id=stream_reader_started._id,
        )

        assert partition_session.id not in stream_reader_started._partition_sessions
        stream_reader_started._partition_sessions[partition_session.id] = partition_session

        return stream_reader_started._partition_sessions[partition_session.id]

    @pytest.fixture()
    def second_partition_session(self, default_reader_settings, stream_reader_started: ReaderStream):
        partition_session = datatypes.PartitionSession(
            id=12,
            topic_path=default_reader_settings.topic,
            partition_id=10,
            state=datatypes.PartitionSession.State.Active,
            committed_offset=self.second_partition_session_offset,
            reader_reconnector_id=self.default_reader_reconnector_id,
            reader_stream_id=stream_reader_started._id,
        )

        assert partition_session.id not in stream_reader_started._partition_sessions
        stream_reader_started._partition_sessions[partition_session.id] = partition_session

        return stream_reader_started._partition_sessions[partition_session.id]

    async def get_started_reader(self, stream, *args, **kwargs) -> ReaderStream:
        reader = ReaderStream(self.default_reader_reconnector_id, *args, **kwargs)
        init_message = object()

        # noinspection PyTypeChecker
        start = asyncio.create_task(reader._start(stream, init_message))

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.InitResponse(session_id="test-session"),
            )
        )

        init_request = await wait_for_fast(stream.from_client.get())
        assert init_request.client_message == init_message

        read_request = await wait_for_fast(stream.from_client.get())
        assert isinstance(read_request.client_message, StreamReadMessage.ReadRequest)

        await start

        await asyncio.sleep(0)
        with pytest.raises(asyncio.QueueEmpty):
            stream.from_client.get_nowait()

        return reader

    @pytest.fixture()
    async def stream_reader_started(self, stream, default_reader_settings) -> ReaderStream:
        return await self.get_started_reader(stream, default_reader_settings)

    @pytest.fixture()
    async def stream_reader(self, stream_reader_started: ReaderStream):
        yield stream_reader_started

        assert stream_reader_started._get_first_error() is None
        await stream_reader_started.close(False)

    @pytest.fixture()
    async def stream_reader_finish_with_error(self, stream_reader_started: ReaderStream):
        yield stream_reader_started

        assert stream_reader_started._get_first_error() is not None
        await stream_reader_started.close(False)

    @staticmethod
    def create_message(
        partition_session: typing.Optional[datatypes.PartitionSession],
        seqno: int,
        offset_delta: int,
    ):
        return PublicMessage(
            seqno=seqno,
            created_at=datetime.datetime(2023, 2, 3, 14, 15),
            message_group_id="test-message-group",
            session_metadata={},
            offset=partition_session._next_message_start_commit_offset + offset_delta - 1,
            written_at=datetime.datetime(2023, 2, 3, 14, 16),
            producer_id="test-producer-id",
            data=bytes(),
            metadata_items={},
            _partition_session=partition_session,
            _commit_start_offset=partition_session._next_message_start_commit_offset + offset_delta - 1,
            _commit_end_offset=partition_session._next_message_start_commit_offset + offset_delta,
        )

    async def send_message(self, stream_reader, message: PublicMessage, new_batch=True):
        await self.send_batch(stream_reader, [message], new_batch=new_batch)

    async def send_batch(self, stream_reader, batch: typing.List[PublicMessage], new_batch=True):
        if len(batch) == 0:
            return

        first_message = batch[0]
        for message in batch:
            assert message._partition_session is first_message._partition_session

        partition_session_id = first_message._partition_session.id

        def batch_count():
            return len(stream_reader._message_batches)

        def batch_size():
            return len(stream_reader._message_batches[partition_session_id].messages)

        initial_batches = batch_count()
        initial_batch_size = 0 if new_batch else batch_size()

        stream = stream_reader._stream  # type: StreamMock
        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.ReadResponse(
                    partition_data=[
                        StreamReadMessage.ReadResponse.PartitionData(
                            partition_session_id=first_message._partition_session.id,
                            batches=[
                                StreamReadMessage.ReadResponse.Batch(
                                    message_data=[
                                        StreamReadMessage.ReadResponse.MessageData(
                                            offset=message.offset,
                                            seq_no=message.seqno,
                                            created_at=message.created_at,
                                            data=message.data,
                                            metadata_items={},
                                            uncompresed_size=len(message.data),
                                            message_group_id=message.message_group_id,
                                        )
                                        for message in batch
                                    ],
                                    producer_id=first_message.producer_id,
                                    write_session_meta=first_message.session_metadata,
                                    codec=Codec.CODEC_RAW,
                                    written_at=first_message.written_at,
                                )
                            ],
                        )
                    ],
                    bytes_size=self.default_batch_size,
                ),
            )
        )
        if new_batch:
            await wait_condition(lambda: batch_count() > initial_batches)
        else:
            await wait_condition(lambda: batch_size() > initial_batch_size)

    async def test_unknown_error(self, stream, stream_reader_finish_with_error):
        class TestError(Exception):
            pass

        test_err = TestError()
        stream.from_server.put_nowait(test_err)

        with pytest.raises(TestError):
            await wait_for_fast(stream_reader_finish_with_error.wait_messages())

        with pytest.raises(TestError):
            stream_reader_finish_with_error.receive_batch_nowait()

    @pytest.mark.parametrize(
        "commit,send_range",
        [
            (
                OffsetsRange(
                    partition_session_committed_offset,
                    partition_session_committed_offset + 1,
                ),
                True,
            ),
            (
                OffsetsRange(
                    partition_session_committed_offset - 1,
                    partition_session_committed_offset,
                ),
                False,
            ),
        ],
    )
    async def test_send_commit_messages(
        self,
        stream,
        stream_reader: ReaderStream,
        partition_session,
        commit: OffsetsRange,
        send_range: bool,
    ):
        @dataclass
        class Commitable(datatypes.ICommittable):
            start: int
            end: int

            def _commit_get_partition_session(self) -> datatypes.PartitionSession:
                return partition_session

            def _commit_get_offsets_range(self) -> OffsetsRange:
                return OffsetsRange(self.start, self.end)

        start_ack_waiters = partition_session._ack_waiters.copy()

        waiter = stream_reader.commit(Commitable(commit.start, commit.end))

        async def wait_message():
            return await wait_for_fast(stream.from_client.get(), timeout=0)

        if send_range:
            msg = await wait_message()  # type: StreamReadMessage.FromClient
            assert msg.client_message == StreamReadMessage.CommitOffsetRequest(
                commit_offsets=[
                    StreamReadMessage.CommitOffsetRequest.PartitionCommitOffset(
                        partition_session_id=partition_session.id,
                        offsets=[commit],
                    )
                ]
            )
            assert partition_session._ack_waiters[-1].end_offset == commit.end
        else:
            assert waiter.future.done()

            with pytest.raises(WaitConditionError):
                msg = await wait_message()
                pass
            assert start_ack_waiters == partition_session._ack_waiters

    async def test_commit_ack_received(self, stream_reader, stream, partition_session, second_partition_session):
        offset1 = self.partition_session_committed_offset + 1
        waiter1 = partition_session.add_waiter(offset1)

        offset2 = self.second_partition_session_offset + 2
        waiter2 = second_partition_session.add_waiter(offset2)

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.CommitOffsetResponse(
                    partitions_committed_offsets=[
                        StreamReadMessage.CommitOffsetResponse.PartitionCommittedOffset(
                            partition_session_id=partition_session.id,
                            committed_offset=offset1,
                        ),
                        StreamReadMessage.CommitOffsetResponse.PartitionCommittedOffset(
                            partition_session_id=second_partition_session.id,
                            committed_offset=offset2,
                        ),
                    ]
                ),
            )
        )

        await wait_for_fast(waiter1.future)
        await wait_for_fast(waiter2.future)

    async def test_close_ack_waiters_when_close_stream_reader(
        self, stream_reader_started: ReaderStream, partition_session
    ):
        waiter = partition_session.add_waiter(self.partition_session_committed_offset + 1)
        await wait_for_fast(stream_reader_started.close(False))

        with pytest.raises(topic_reader_asyncio.PublicTopicReaderPartitionExpiredError):
            waiter.future.result()

    async def test_flush(self, stream, stream_reader_started: ReaderStream, partition_session):
        offset = self.partition_session_committed_offset + 1
        waiter = partition_session.add_waiter(offset)

        with pytest.raises(WaitConditionError):
            await wait_for_fast(stream_reader_started.flush(), timeout=0.1)

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.CommitOffsetResponse(
                    partitions_committed_offsets=[
                        StreamReadMessage.CommitOffsetResponse.PartitionCommittedOffset(
                            partition_session_id=partition_session.id,
                            committed_offset=offset,
                        ),
                    ]
                ),
            )
        )

        await stream_reader_started.flush()
        # don't raises
        assert waiter.future.result() is None

        await wait_for_fast(stream_reader_started.close(False))

    async def test_commit_ranges_for_received_messages(
        self, stream, stream_reader_started: ReaderStream, partition_session
    ):
        m1 = self.create_message(partition_session, 1, 1)
        m2 = self.create_message(partition_session, 2, 10)
        m2._commit_start_offset = m1.offset + 1

        await self.send_message(stream_reader_started, m1)
        await self.send_message(stream_reader_started, m2, new_batch=False)

        await stream_reader_started.wait_messages()
        received = stream_reader_started.receive_batch_nowait().messages
        assert received == [m1, m2]

        await stream_reader_started.close(False)

    # noinspection PyTypeChecker
    @pytest.mark.parametrize(
        "batch,data_out",
        [
            (
                PublicBatch(
                    messages=[
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            data=rb"123",
                            metadata_items={},
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        )
                    ],
                    _partition_session=None,
                    _bytes_size=0,
                    _codec=Codec.CODEC_RAW,
                ),
                [bytes(rb"123")],
            ),
            (
                PublicBatch(
                    messages=[
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            data=gzip.compress(rb"123"),
                            metadata_items={},
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        )
                    ],
                    _partition_session=None,
                    _bytes_size=0,
                    _codec=Codec.CODEC_GZIP,
                ),
                [bytes(rb"123")],
            ),
            (
                PublicBatch(
                    messages=[
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            metadata_items={},
                            data=rb"123",
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        ),
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            data=rb"456",
                            metadata_items={},
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        ),
                    ],
                    _partition_session=None,
                    _bytes_size=0,
                    _codec=Codec.CODEC_RAW,
                ),
                [bytes(rb"123"), bytes(rb"456")],
            ),
            (
                PublicBatch(
                    messages=[
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            data=gzip.compress(rb"123"),
                            metadata_items={},
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        ),
                        PublicMessage(
                            seqno=1,
                            created_at=datetime.datetime(2023, 3, 14, 15, 41),
                            message_group_id="",
                            session_metadata={},
                            offset=1,
                            written_at=datetime.datetime(2023, 3, 14, 15, 42),
                            producer_id="asd",
                            data=gzip.compress(rb"456"),
                            metadata_items={},
                            _partition_session=None,
                            _commit_start_offset=5,
                            _commit_end_offset=15,
                        ),
                    ],
                    _partition_session=None,
                    _bytes_size=0,
                    _codec=Codec.CODEC_GZIP,
                ),
                [bytes(rb"123"), bytes(rb"456")],
            ),
        ],
    )
    async def test_decode_loop(self, stream_reader, batch: PublicBatch, data_out: typing.List[bytes]):
        assert len(batch.messages) == len(data_out)

        expected = copy.deepcopy(batch)
        expected._codec = Codec.CODEC_RAW

        for index, data in enumerate(data_out):
            expected.messages[index].data = data

        await wait_for_fast(stream_reader._decode_batch_inplace(batch))

        assert batch == expected

    async def test_error_from_status_code(self, stream, stream_reader_finish_with_error):
        # noinspection PyTypeChecker
        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(
                    status=issues.StatusCode.OVERLOADED,
                    issues=[],
                ),
                server_message=None,
            )
        )

        with pytest.raises(issues.Overloaded):
            await wait_for_fast(stream_reader_finish_with_error.wait_messages())

        with pytest.raises(issues.Overloaded):
            stream_reader_finish_with_error.receive_batch_nowait()

    async def test_init_reader(self, stream, default_reader_settings):
        reader = ReaderStream(self.default_reader_reconnector_id, default_reader_settings)
        init_message = StreamReadMessage.InitRequest(
            consumer="test-consumer",
            topics_read_settings=[
                StreamReadMessage.InitRequest.TopicReadSettings(
                    path="/local/test-topic",
                    partition_ids=[],
                    max_lag=None,
                    read_from=None,
                )
            ],
            auto_partitioning_support=False,
        )
        start_task = asyncio.create_task(reader._start(stream, init_message))

        sent_message = await wait_for_fast(stream.from_client.get())
        expected_sent_init_message = StreamReadMessage.FromClient(client_message=init_message)
        assert sent_message == expected_sent_init_message

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.InitResponse(session_id="test"),
            )
        )

        await start_task

        read_request = await wait_for_fast(stream.from_client.get())
        assert read_request.client_message == StreamReadMessage.ReadRequest(
            bytes_size=default_reader_settings.buffer_size_bytes,
        )

        assert reader._session_id == "test"
        await reader.close(False)

    async def test_start_partition(
        self,
        stream_reader: ReaderStream,
        stream,
        default_reader_settings,
        partition_session,
    ):
        def session_count():
            return len(stream_reader._partition_sessions)

        initial_session_count = session_count()

        test_partition_id = partition_session.partition_id + 1
        test_partition_session_id = partition_session.id + 1
        test_topic_path = default_reader_settings.topic + "-asd"
        test_partition_committed_offset = 18

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.StartPartitionSessionRequest(
                    partition_session=StreamReadMessage.PartitionSession(
                        partition_session_id=test_partition_session_id,
                        path=test_topic_path,
                        partition_id=test_partition_id,
                    ),
                    committed_offset=test_partition_committed_offset,
                    partition_offsets=OffsetsRange(
                        start=0,
                        end=0,
                    ),
                ),
            ),
        )
        response = await wait_for_fast(stream.from_client.get())
        assert response == StreamReadMessage.FromClient(
            client_message=StreamReadMessage.StartPartitionSessionResponse(
                partition_session_id=test_partition_session_id,
                read_offset=None,
                commit_offset=None,
            )
        )

        assert len(stream_reader._partition_sessions) == initial_session_count + 1
        assert stream_reader._partition_sessions[test_partition_session_id] == datatypes.PartitionSession(
            id=test_partition_session_id,
            state=datatypes.PartitionSession.State.Active,
            topic_path=test_topic_path,
            partition_id=test_partition_id,
            committed_offset=test_partition_committed_offset,
            reader_reconnector_id=self.default_reader_reconnector_id,
            reader_stream_id=stream_reader._id,
        )

    async def test_partition_stop_force(self, stream, stream_reader, partition_session):
        def session_count():
            return len(stream_reader._partition_sessions)

        initial_session_count = session_count()

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.StopPartitionSessionRequest(
                    partition_session_id=partition_session.id,
                    graceful=False,
                    committed_offset=0,
                ),
            )
        )

        await asyncio.sleep(0)  # wait next loop
        with pytest.raises(asyncio.QueueEmpty):
            stream.from_client.get_nowait()

        assert session_count() == initial_session_count - 1
        assert partition_session.id not in stream_reader._partition_sessions

    async def test_partition_stop_graceful(self, stream, stream_reader, partition_session):
        def session_count():
            return len(stream_reader._partition_sessions)

        initial_session_count = session_count()

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.StopPartitionSessionRequest(
                    partition_session_id=partition_session.id,
                    graceful=True,
                    committed_offset=0,
                ),
            )
        )

        resp = await wait_for_fast(stream.from_client.get())  # type: StreamReadMessage.FromClient
        assert session_count() == initial_session_count - 1
        assert partition_session.id not in stream_reader._partition_sessions
        assert resp.client_message == StreamReadMessage.StopPartitionSessionResponse(
            partition_session_id=partition_session.id
        )

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.StopPartitionSessionRequest(
                    partition_session_id=partition_session.id,
                    graceful=False,
                    committed_offset=0,
                ),
            )
        )

        await asyncio.sleep(0)  # wait next loop
        with pytest.raises(asyncio.QueueEmpty):
            stream.from_client.get_nowait()

    async def test_end_partition_session(self, stream, stream_reader, partition_session):
        def session_count():
            return len(stream_reader._partition_sessions)

        initial_session_count = session_count()

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.EndPartitionSession(
                    partition_session_id=partition_session.id,
                    adjacent_partition_ids=[],
                    child_partition_ids=[20, 30],
                ),
            )
        )

        await asyncio.sleep(0)  # wait next loop
        with pytest.raises(asyncio.QueueEmpty):
            stream.from_client.get_nowait()

        assert session_count() == initial_session_count
        assert partition_session.id in stream_reader._partition_sessions
        assert partition_session.ended

    @pytest.mark.parametrize(
        "graceful",
        (
            [True],
            [False],
        ),
    )
    async def test_free_buffer_after_partition_stop(self, stream, stream_reader, partition_session, graceful):
        initial_buffer_size = stream_reader._buffer_size_bytes
        message_size = initial_buffer_size - 1

        t = datetime.datetime.now(datetime.timezone.utc)

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(issues.StatusCode.SUCCESS, []),
                server_message=StreamReadMessage.ReadResponse(
                    bytes_size=message_size,
                    partition_data=[
                        StreamReadMessage.ReadResponse.PartitionData(
                            partition_session_id=partition_session.id,
                            batches=[
                                StreamReadMessage.ReadResponse.Batch(
                                    message_data=[
                                        StreamReadMessage.ReadResponse.MessageData(
                                            partition_session.committed_offset + 1,
                                            seq_no=123,
                                            created_at=t,
                                            data=bytes(),
                                            metadata_items={},
                                            uncompresed_size=message_size,
                                            message_group_id="test-message-group",
                                        )
                                    ],
                                    producer_id="asd",
                                    write_session_meta={},
                                    codec=Codec.CODEC_RAW,
                                    written_at=t,
                                )
                            ],
                        )
                    ],
                ),
            )
        )

        def message_received():
            return len(stream_reader._message_batches) > 0

        await wait_condition(message_received)

        assert stream_reader._buffer_size_bytes == initial_buffer_size - message_size

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(issues.StatusCode.SUCCESS, []),
                server_message=StreamReadMessage.StopPartitionSessionRequest(
                    partition_session_id=partition_session.id,
                    graceful=graceful,
                    committed_offset=partition_session.committed_offset,
                ),
            )
        )

        await wait_condition(lambda: partition_session.closed)

        batch = stream_reader.receive_batch_nowait()
        assert not batch.alive
        assert stream_reader._buffer_size_bytes == initial_buffer_size

    async def test_receive_message_from_server(
        self,
        stream_reader,
        stream,
        partition_session: datatypes.PartitionSession,
        second_partition_session,
    ):
        def reader_batch_count():
            return len(stream_reader._message_batches)

        initial_buffer_size = stream_reader._buffer_size_bytes
        initial_batch_count = reader_batch_count()

        bytes_size = 10
        created_at = datetime.datetime(2020, 1, 1, 18, 12)
        written_at = datetime.datetime(2023, 2, 1, 18, 12)
        producer_id = "test-producer-id"
        data = "123".encode()
        session_meta = {"a": "b"}
        message_group_id = "test-message-group-id"

        expected_message_offset = partition_session.committed_offset

        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.ReadResponse(
                    bytes_size=bytes_size,
                    partition_data=[
                        StreamReadMessage.ReadResponse.PartitionData(
                            partition_session_id=partition_session.id,
                            batches=[
                                StreamReadMessage.ReadResponse.Batch(
                                    message_data=[
                                        StreamReadMessage.ReadResponse.MessageData(
                                            offset=expected_message_offset,
                                            seq_no=2,
                                            created_at=created_at,
                                            data=data,
                                            uncompresed_size=len(data),
                                            metadata_items={},
                                            message_group_id=message_group_id,
                                        )
                                    ],
                                    producer_id=producer_id,
                                    write_session_meta=session_meta,
                                    codec=Codec.CODEC_RAW,
                                    written_at=written_at,
                                )
                            ],
                        )
                    ],
                ),
            )
        ),

        await wait_condition(lambda: reader_batch_count() == initial_batch_count + 1)

        assert stream_reader._buffer_size_bytes == initial_buffer_size - bytes_size

        _, last_batch = stream_reader._message_batches.popitem()
        assert last_batch == PublicBatch(
            messages=[
                PublicMessage(
                    seqno=2,
                    created_at=created_at,
                    message_group_id=message_group_id,
                    session_metadata=session_meta,
                    offset=expected_message_offset,
                    written_at=written_at,
                    producer_id=producer_id,
                    data=data,
                    metadata_items={},
                    _partition_session=partition_session,
                    _commit_start_offset=expected_message_offset,
                    _commit_end_offset=expected_message_offset + 1,
                )
            ],
            _partition_session=partition_session,
            _bytes_size=bytes_size,
            _codec=Codec.CODEC_RAW,
        )

    async def test_read_batches(self, stream_reader, partition_session, second_partition_session):
        created_at = datetime.datetime(2020, 2, 1, 18, 12)
        created_at2 = datetime.datetime(2020, 2, 2, 18, 12)
        created_at3 = datetime.datetime(2020, 2, 3, 18, 12)
        created_at4 = datetime.datetime(2020, 2, 4, 18, 12)
        written_at = datetime.datetime(2023, 3, 1, 18, 12)
        written_at2 = datetime.datetime(2023, 3, 2, 18, 12)
        producer_id = "test-producer-id"
        producer_id2 = "test-producer-id"
        data = "123".encode()
        data2 = "1235".encode()
        session_meta = {"a": "b"}
        session_meta2 = {"b": "c"}

        message_group_id = "test-message-group-id"
        message_group_id2 = "test-message-group-id-2"

        partition1_mess1_expected_offset = partition_session.committed_offset
        partition2_mess1_expected_offset = second_partition_session.committed_offset
        partition2_mess2_expected_offset = second_partition_session.committed_offset + 1
        partition2_mess3_expected_offset = second_partition_session.committed_offset + 2

        batches = stream_reader._read_response_to_batches(
            StreamReadMessage.ReadResponse(
                bytes_size=3,
                partition_data=[
                    StreamReadMessage.ReadResponse.PartitionData(
                        partition_session_id=partition_session.id,
                        batches=[
                            StreamReadMessage.ReadResponse.Batch(
                                message_data=[
                                    StreamReadMessage.ReadResponse.MessageData(
                                        offset=partition1_mess1_expected_offset,
                                        seq_no=3,
                                        created_at=created_at,
                                        data=data,
                                        metadata_items={},
                                        uncompresed_size=len(data),
                                        message_group_id=message_group_id,
                                    )
                                ],
                                producer_id=producer_id,
                                write_session_meta=session_meta,
                                codec=Codec.CODEC_RAW,
                                written_at=written_at,
                            )
                        ],
                    ),
                    StreamReadMessage.ReadResponse.PartitionData(
                        partition_session_id=second_partition_session.id,
                        batches=[
                            StreamReadMessage.ReadResponse.Batch(
                                message_data=[
                                    StreamReadMessage.ReadResponse.MessageData(
                                        offset=partition2_mess1_expected_offset,
                                        seq_no=2,
                                        created_at=created_at2,
                                        data=data,
                                        metadata_items={},
                                        uncompresed_size=len(data),
                                        message_group_id=message_group_id,
                                    )
                                ],
                                producer_id=producer_id,
                                write_session_meta=session_meta,
                                codec=Codec.CODEC_RAW,
                                written_at=written_at2,
                            ),
                            StreamReadMessage.ReadResponse.Batch(
                                message_data=[
                                    StreamReadMessage.ReadResponse.MessageData(
                                        offset=partition2_mess2_expected_offset,
                                        seq_no=3,
                                        created_at=created_at3,
                                        data=data2,
                                        metadata_items={},
                                        uncompresed_size=len(data2),
                                        message_group_id=message_group_id,
                                    ),
                                    StreamReadMessage.ReadResponse.MessageData(
                                        offset=partition2_mess3_expected_offset,
                                        seq_no=5,
                                        created_at=created_at4,
                                        data=data,
                                        metadata_items={},
                                        uncompresed_size=len(data),
                                        message_group_id=message_group_id2,
                                    ),
                                ],
                                producer_id=producer_id2,
                                write_session_meta=session_meta2,
                                codec=Codec.CODEC_RAW,
                                written_at=written_at2,
                            ),
                        ],
                    ),
                ],
            )
        )

        last0 = batches[0]
        last1 = batches[1]
        last2 = batches[2]

        assert last0 == PublicBatch(
            messages=[
                PublicMessage(
                    seqno=3,
                    created_at=created_at,
                    message_group_id=message_group_id,
                    session_metadata=session_meta,
                    offset=partition1_mess1_expected_offset,
                    written_at=written_at,
                    producer_id=producer_id,
                    data=data,
                    metadata_items={},
                    _partition_session=partition_session,
                    _commit_start_offset=partition1_mess1_expected_offset,
                    _commit_end_offset=partition1_mess1_expected_offset + 1,
                )
            ],
            _partition_session=partition_session,
            _bytes_size=1,
            _codec=Codec.CODEC_RAW,
        )
        assert last1 == PublicBatch(
            messages=[
                PublicMessage(
                    seqno=2,
                    created_at=created_at2,
                    message_group_id=message_group_id,
                    session_metadata=session_meta,
                    offset=partition2_mess1_expected_offset,
                    written_at=written_at2,
                    producer_id=producer_id,
                    data=data,
                    metadata_items={},
                    _partition_session=second_partition_session,
                    _commit_start_offset=partition2_mess1_expected_offset,
                    _commit_end_offset=partition2_mess1_expected_offset + 1,
                )
            ],
            _partition_session=second_partition_session,
            _bytes_size=1,
            _codec=Codec.CODEC_RAW,
        )
        assert last2 == PublicBatch(
            messages=[
                PublicMessage(
                    seqno=3,
                    created_at=created_at3,
                    message_group_id=message_group_id,
                    session_metadata=session_meta2,
                    offset=partition2_mess2_expected_offset,
                    written_at=written_at2,
                    producer_id=producer_id2,
                    data=data2,
                    metadata_items={},
                    _partition_session=second_partition_session,
                    _commit_start_offset=partition2_mess2_expected_offset,
                    _commit_end_offset=partition2_mess2_expected_offset + 1,
                ),
                PublicMessage(
                    seqno=5,
                    created_at=created_at4,
                    message_group_id=message_group_id2,
                    session_metadata=session_meta2,
                    offset=partition2_mess3_expected_offset,
                    written_at=written_at2,
                    producer_id=producer_id,
                    data=data,
                    metadata_items={},
                    _partition_session=second_partition_session,
                    _commit_start_offset=partition2_mess3_expected_offset,
                    _commit_end_offset=partition2_mess3_expected_offset + 1,
                ),
            ],
            _partition_session=second_partition_session,
            _bytes_size=1,
            _codec=Codec.CODEC_RAW,
        )

    @pytest.mark.parametrize(
        "batches_before,expected_message,batches_after",
        [
            ({}, None, {}),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    )
                },
                stub_message(1),
                {},
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1), stub_message(2)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                stub_message(1),
                {
                    0: PublicBatch(
                        messages=[stub_message(2)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(2), stub_message(3)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                stub_message(1),
                {
                    1: PublicBatch(
                        messages=[stub_message(2), stub_message(3)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    )
                },
            ),
        ],
    )
    async def test_read_message(
        self,
        stream_reader,
        batches_before: typing.List[datatypes.PublicBatch],
        expected_message: PublicMessage,
        batches_after: typing.List[datatypes.PublicBatch],
    ):
        stream_reader._message_batches = OrderedDict(batches_before)
        mess = stream_reader.receive_message_nowait()

        assert mess == expected_message
        assert dict(stream_reader._message_batches) == batches_after

    @pytest.mark.parametrize(
        "batches,expected_order",
        [
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1)],
                        _partition_session=stub_partition_session(0, ended=True),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    )
                },
                [1],
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1), stub_message(2)],
                        _partition_session=stub_partition_session(0, ended=True),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                [1, 2, 3, 4],
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1), stub_message(2)],
                        _partition_session=stub_partition_session(0),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1, ended=True),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                    2: PublicBatch(
                        messages=[stub_message(5)],
                        _partition_session=stub_partition_session(2),
                        _bytes_size=0,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                [1, 3, 4, 5, 2],
            ),
        ],
    )
    async def test_read_message_autosplit_order(
        self,
        stream_reader,
        batches: typing.Dict[int, datatypes.PublicBatch],
        expected_order: typing.List[int],
    ):
        stream_reader._message_batches = OrderedDict(batches)

        for id, batch in batches.items():
            ps = batch._partition_session
            stream_reader._partition_sessions[id] = ps

        result = []
        for _ in range(len(expected_order)):
            mess = stream_reader.receive_message_nowait()
            result.append(mess.data)

        assert result == expected_order
        assert stream_reader.receive_message_nowait() is None

    @pytest.mark.parametrize(
        "batches_before,max_messages,actual_messages,batches_after",
        [
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    )
                },
                None,
                1,
                {},
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1), stub_message(2)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                1,
                1,
                {
                    1: PublicBatch(
                        messages=[stub_message(3), stub_message(4)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    ),
                    0: PublicBatch(
                        messages=[stub_message(2)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=2,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
            ),
            (
                {
                    0: PublicBatch(
                        messages=[stub_message(1)],
                        _partition_session=stub_partition_session(),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    ),
                    1: PublicBatch(
                        messages=[stub_message(2), stub_message(3)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    ),
                },
                100,
                1,
                {
                    1: PublicBatch(
                        messages=[stub_message(2), stub_message(3)],
                        _partition_session=stub_partition_session(1),
                        _bytes_size=4,
                        _codec=Codec.CODEC_RAW,
                    )
                },
            ),
        ],
    )
    async def test_read_batch_max_messages(
        self,
        stream_reader,
        batches_before: typing.List[datatypes.PublicBatch],
        max_messages: typing.Optional[int],
        actual_messages: int,
        batches_after: typing.List[datatypes.PublicBatch],
    ):
        stream_reader._message_batches = OrderedDict(batches_before)
        batch = stream_reader.receive_batch_nowait(max_messages=max_messages)

        assert len(batch.messages) == actual_messages
        assert stream_reader._message_batches == OrderedDict(batches_after)

    async def test_receive_batch_nowait(self, stream, stream_reader, partition_session):
        assert stream_reader.receive_batch_nowait() is None

        initial_buffer_size = stream_reader._buffer_size_bytes

        mess1 = self.create_message(partition_session, 1, 1)
        await self.send_message(stream_reader, mess1)

        mess2 = self.create_message(partition_session, 2, 1)
        await self.send_message(stream_reader, mess2, new_batch=False)

        assert stream_reader._buffer_size_bytes == initial_buffer_size - 2 * self.default_batch_size

        received = stream_reader.receive_batch_nowait()
        assert received == PublicBatch(
            messages=[mess1, mess2],
            _partition_session=mess1._partition_session,
            _bytes_size=self.default_batch_size * 2,
            _codec=Codec.CODEC_RAW,
        )

        assert stream_reader._buffer_size_bytes == initial_buffer_size

        assert (
            StreamReadMessage.ReadRequest(self.default_batch_size * 2) == stream.from_client.get_nowait().client_message
        )

        with pytest.raises(asyncio.QueueEmpty):
            stream.from_client.get_nowait()

    async def test_receive_message_nowait(self, stream, stream_reader, partition_session):
        assert stream_reader.receive_batch_nowait() is None

        initial_buffer_size = stream_reader._buffer_size_bytes

        await self.send_batch(
            stream_reader,
            [
                self.create_message(partition_session, 1, 1),
                self.create_message(partition_session, 2, 1),
            ],
        )
        await self.send_batch(
            stream_reader,
            [
                self.create_message(partition_session, 10, 1),
            ],
            new_batch=False,
        )

        assert stream_reader._buffer_size_bytes == initial_buffer_size - 2 * self.default_batch_size

        for expected_seqno in [1, 2, 10]:
            mess = stream_reader.receive_message_nowait()
            assert mess.seqno == expected_seqno

        assert stream_reader._buffer_size_bytes == initial_buffer_size

    async def test_update_token(self, stream):
        settings = PublicReaderSettings(
            consumer="test-consumer",
            topic="test-topic",
            update_token_interval=0.1,
        )
        reader = await self.get_started_reader(stream, settings, get_token_function=lambda: "foo-bar")

        assert stream.from_client.empty()

        expected = StreamReadMessage.FromClient(UpdateTokenRequest(token="foo-bar"))
        got = await wait_for_fast(stream.from_client.get())
        assert expected == got, "send update token request"

        await asyncio.sleep(0.2)
        assert stream.from_client.empty(), "no answer - no new update request"

        await stream.from_server.put(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=UpdateTokenResponse(),
            )
        )

        got = await wait_for_fast(stream.from_client.get())
        assert expected == got

        await reader.close(False)

    async def test_read_unknown_message(self, stream, stream_reader, caplog):
        class TestMessage:
            pass

        # noinspection PyTypeChecker
        stream.from_server.put_nowait(
            StreamReadMessage.FromServer(
                server_status=ServerStatus(
                    status=issues.StatusCode.SUCCESS,
                    issues=[],
                ),
                server_message=TestMessage(),
            )
        )

        def logged():
            for rec in caplog.records:
                if TestMessage.__name__ in rec.message:
                    return True
            return False

        await wait_condition(logged)

    async def test_init_timeout_parameter(self, stream, default_reader_settings):
        """Test that ReaderStream._start calls stream.receive with timeout=10"""
        reader = ReaderStream(self.default_reader_reconnector_id, default_reader_settings)
        init_message = default_reader_settings._init_message()

        # Mock stream.receive to check if timeout is passed
        with mock.patch.object(stream, "receive") as mock_receive:
            mock_receive.return_value = StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.InitResponse(session_id="test_session"),
            )

            await reader._start(stream, init_message)

            # Verify that receive was called with timeout
            mock_receive.assert_called_with(timeout=DEFAULT_INITIAL_RESPONSE_TIMEOUT)

        await reader.close(False)

    async def test_init_timeout_behavior(self, stream, default_reader_settings):
        """Test that ReaderStream._start raises TopicReaderError when receive times out"""
        reader = ReaderStream(self.default_reader_reconnector_id, default_reader_settings)
        init_message = default_reader_settings._init_message()

        # Mock stream.receive to directly raise TimeoutError when called with timeout
        async def timeout_receive(timeout=None):
            if timeout == DEFAULT_INITIAL_RESPONSE_TIMEOUT:
                raise asyncio.TimeoutError("Simulated timeout")
            return StreamReadMessage.FromServer(
                server_status=ServerStatus(ydb_status_codes_pb2.StatusIds.SUCCESS, []),
                server_message=StreamReadMessage.InitResponse(session_id="test_session"),
            )

        with mock.patch.object(stream, "receive", side_effect=timeout_receive):
            # Should raise TopicReaderError with timeout message
            with pytest.raises(TopicReaderError, match="Timeout waiting for init response"):
                await reader._start(stream, init_message)

        await reader.close(False)


@pytest.mark.asyncio
class TestReaderReconnector:
    async def test_reconnect_on_repeatable_error(self, monkeypatch):
        test_error = issues.Overloaded("test error")

        async def wait_error():
            raise test_error

        reader_stream_mock_with_error = mock.Mock(ReaderStream)
        reader_stream_mock_with_error._id = 0
        reader_stream_mock_with_error.wait_error = mock.AsyncMock(side_effect=wait_error)

        async def wait_messages_with_error():
            raise test_error

        reader_stream_mock_with_error.wait_messages = mock.AsyncMock(side_effect=wait_messages_with_error)

        async def wait_forever():
            f = asyncio.Future()
            await f

        reader_stream_with_messages = mock.Mock(ReaderStream)
        reader_stream_with_messages._id = 0
        reader_stream_with_messages.wait_error = mock.AsyncMock(side_effect=wait_forever)
        reader_stream_with_messages.wait_messages.return_value = None

        stream_index = 0

        async def stream_create(
            reader_reconnector_id: int,
            driver: SupportedDriverType,
            settings: PublicReaderSettings,
        ):
            nonlocal stream_index
            stream_index += 1
            if stream_index == 1:
                return reader_stream_mock_with_error
            elif stream_index == 2:
                return reader_stream_with_messages
            else:
                raise Exception("unexpected create stream")

        with mock.patch.object(ReaderStream, "create", stream_create):
            reconnector = ReaderReconnector(mock.Mock(), PublicReaderSettings("", ""))
            await wait_for_fast(reconnector.wait_message())

        reader_stream_mock_with_error.wait_error.assert_any_await()
        reader_stream_mock_with_error.wait_messages.assert_any_await()



import datetime
import typing
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Optional, List, Union, Dict

# Workaround for good IDE and universal for runtime
if typing.TYPE_CHECKING:
    from ..v4.protos import ydb_topic_pb2
else:
    from ..common.protos import ydb_topic_pb2

from .common_utils import IToProto
from ...scheme import SchemeEntry


@dataclass
# need similar struct to PublicDescribeTopicResult
class CreateTopicRequestParams:
    path: str
    min_active_partitions: Optional[int]
    max_active_partitions: Optional[int]
    partition_count_limit: Optional[int]
    retention_period: Optional[datetime.timedelta]
    retention_storage_mb: Optional[int]
    supported_codecs: Optional[List[Union["PublicCodec", int]]]
    partition_write_speed_bytes_per_second: Optional[int]
    partition_write_burst_bytes: Optional[int]
    attributes: Optional[Dict[str, str]]
    consumers: Optional[List[Union["PublicConsumer", str]]]
    metering_mode: Optional["PublicMeteringMode"]
    auto_partitioning_settings: Optional["PublicAutoPartitioningSettings"]


@dataclass
class AlterTopicRequestParams:
    path: str
    set_min_active_partitions: Optional[int]
    set_max_active_partitions: Optional[int]
    set_partition_count_limit: Optional[int]
    add_consumers: Optional[List[Union["PublicConsumer", str]]]
    alter_consumers: Optional[List[Union["PublicAlterConsumer", str]]]
    drop_consumers: Optional[List[str]]
    alter_attributes: Optional[Dict[str, str]]
    set_metering_mode: Optional["PublicMeteringMode"]
    set_partition_write_speed_bytes_per_second: Optional[int]
    set_partition_write_burst_bytes: Optional[int]
    set_retention_period: Optional[datetime.timedelta]
    set_retention_storage_mb: Optional[int]
    set_supported_codecs: Optional[List[Union["PublicCodec", int]]]
    alter_auto_partitioning_settings: Optional["PublicAlterAutoPartitioningSettings"]


class PublicCodec(int):
    """
    Codec value may contain any int number.

    Values below is only well-known predefined values,
    but protocol support custom codecs.
    """

    UNSPECIFIED = 0
    RAW = 1
    GZIP = 2
    LZOP = 3  # Has not supported codec in standard library
    ZSTD = 4  # Has not supported codec in standard library


class PublicMeteringMode(IntEnum):
    UNSPECIFIED = 0
    RESERVED_CAPACITY = 1
    REQUEST_UNITS = 2


class PublicAutoPartitioningStrategy(IntEnum):
    UNSPECIFIED = 0
    DISABLED = 1
    SCALE_UP = 2
    SCALE_UP_AND_DOWN = 3
    PAUSED = 4


@dataclass
class PublicAutoPartitioningSettings:
    strategy: Optional["PublicAutoPartitioningStrategy"] = None
    stabilization_window: Optional[datetime.timedelta] = None
    down_utilization_percent: Optional[int] = None
    up_utilization_percent: Optional[int] = None


@dataclass
class PublicAlterAutoPartitioningSettings:
    set_strategy: Optional["PublicAutoPartitioningStrategy"] = None
    set_stabilization_window: Optional[datetime.timedelta] = None
    set_down_utilization_percent: Optional[int] = None
    set_up_utilization_percent: Optional[int] = None


@dataclass
class PublicConsumer:
    name: str
    important: bool = False
    """
    Consumer may be marked as 'important'. It means messages for this consumer will never expire due to retention.
    User should take care that such consumer never stalls, to prevent running out of disk space.
    """

    read_from: Optional[datetime.datetime] = None
    "All messages with smaller server written_at timestamp will be skipped."

    supported_codecs: List[PublicCodec] = field(default_factory=lambda: list())
    """
    List of supported codecs by this consumer.
    supported_codecs on topic must be contained inside this list.
    """

    attributes: Dict[str, str] = field(default_factory=lambda: dict())
    "Attributes of consumer"

    consumer_stats: Optional["PublicConsumer.ConsumerStats"] = None

    @dataclass
    class ConsumerStats:
        min_partitions_last_read_time: datetime.datetime
        "Minimal timestamp of last read from partitions."

        max_read_time_lag: datetime.timedelta
        """
        Maximum of differences between timestamp of read and write timestamp for all messages,
        read during last minute.
        """

        max_write_time_lag: datetime.timedelta
        """
        Maximum of differences between write timestamp and create timestamp for all messages,
        written during last minute.
        """

        bytes_read: "PublicMultipleWindowsStat"
        "Bytes read statistics."


@dataclass
class PublicAlterConsumer:
    name: str
    set_important: Optional[bool] = None
    """
    Consumer may be marked as 'important'. It means messages for this consumer will never expire due to retention.
    User should take care that such consumer never stalls, to prevent running out of disk space.
    """

    set_read_from: Optional[datetime.datetime] = None
    "All messages with smaller server written_at timestamp will be skipped."

    set_supported_codecs: Optional[List[PublicCodec]] = None
    """
    List of supported codecs by this consumer.
    supported_codecs on topic must be contained inside this list.
    """

    alter_attributes: Optional[Dict[str, str]] = None
    "Attributes of consumer"


@dataclass
class DropTopicRequestParams(IToProto):
    path: str

    def to_proto(self) -> ydb_topic_pb2.DropTopicRequest:
        return ydb_topic_pb2.DropTopicRequest(path=self.path)


@dataclass
class DescribeTopicRequestParams(IToProto):
    path: str
    include_stats: bool

    def to_proto(self) -> ydb_topic_pb2.DescribeTopicRequest:
        return ydb_topic_pb2.DescribeTopicRequest(path=self.path, include_stats=self.include_stats)


@dataclass
# Need similar struct to CreateTopicRequestParams
class PublicDescribeTopicResult:
    self: SchemeEntry
    "Description of scheme object"

    min_active_partitions: int
    "Minimum partition count auto merge would stop working at"

    max_active_partitions: int
    "Minimum partition count auto split would stop working at"

    partition_count_limit: int
    "Limit for total partition count, including active (open for write) and read-only partitions"

    partitions: List["PublicDescribeTopicResult.PartitionInfo"]
    "Partitions description"

    retention_period: datetime.timedelta
    "How long data in partition should be stored"

    retention_storage_mb: int
    "How much data in partition should be stored. Zero value means infinite limit"

    supported_codecs: List[PublicCodec]
    "List of allowed codecs for writers"

    partition_write_speed_bytes_per_second: int
    "Partition write speed in bytes per second"

    partition_write_burst_bytes: int
    "Burst size for write in partition, in bytes"

    attributes: Dict[str, str]
    """User and server attributes of topic. Server attributes starts from "_" and will be validated by server."""

    consumers: List[PublicConsumer]
    """List of consumers for this topic"""

    metering_mode: PublicMeteringMode
    "Metering settings"

    topic_stats: "PublicDescribeTopicResult.TopicStats"
    "Statistics of topic"

    auto_partitioning_settings: "PublicAutoPartitioningSettings"

    @dataclass
    class PartitionInfo:
        partition_id: int
        "Partition identifier"

        active: bool
        "Is partition open for write"

        child_partition_ids: List[int]
        "Ids of partitions which was formed when this partition was split or merged"

        parent_partition_ids: List[int]
        "Ids of partitions from which this partition was formed by split or merge"

        partition_stats: Optional["PublicPartitionStats"]
        "Stats for partition, filled only when include_stats in request is true"

    @dataclass
    class TopicStats:
        store_size_bytes: int
        "Approximate size of topic"

        min_last_write_time: datetime.datetime
        "Minimum of timestamps of last write among all partitions."

        max_write_time_lag: datetime.timedelta
        """
        Maximum of differences between write timestamp and create timestamp for all messages,
        written during last minute.
        """

        bytes_written: "PublicMultipleWindowsStat"
        "How much bytes were written statistics."


@dataclass
class PublicPartitionStats:
    partition_start: int
    "first message offset in the partition"

    partition_end: int
    "offset after last stored message offset in the partition (last offset + 1)"

    store_size_bytes: int
    "Approximate size of partition"

    last_write_time: datetime.datetime
    "Timestamp of last write"

    max_write_time_lag: datetime.timedelta
    "Maximum of differences between write timestamp and create timestamp for all messages, written during last minute."

    bytes_written: "PublicMultipleWindowsStat"
    "How much bytes were written during several windows in this partition."

    partition_node_id: int
    "Host where tablet for this partition works. Useful for debugging purposes."


@dataclass
class PublicMultipleWindowsStat:
    per_minute: int
    per_hour: int
    per_day: int



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: draft/ydb_federated_query_v1.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v5.draft.protos import ydb_federated_query_pb2 as draft_dot_protos_dot_ydb__federated__query__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"draft/ydb_federated_query_v1.proto\x12\x11\x46\x65\x64\x65ratedQuery.V1\x1a&draft/protos/ydb_federated_query.proto2\xd6\x0f\n\x15\x46\x65\x64\x65ratedQueryService\x12V\n\x0b\x43reateQuery\x12\".FederatedQuery.CreateQueryRequest\x1a#.FederatedQuery.CreateQueryResponse\x12V\n\x0bListQueries\x12\".FederatedQuery.ListQueriesRequest\x1a#.FederatedQuery.ListQueriesResponse\x12\\\n\rDescribeQuery\x12$.FederatedQuery.DescribeQueryRequest\x1a%.FederatedQuery.DescribeQueryResponse\x12_\n\x0eGetQueryStatus\x12%.FederatedQuery.GetQueryStatusRequest\x1a&.FederatedQuery.GetQueryStatusResponse\x12V\n\x0bModifyQuery\x12\".FederatedQuery.ModifyQueryRequest\x1a#.FederatedQuery.ModifyQueryResponse\x12V\n\x0b\x44\x65leteQuery\x12\".FederatedQuery.DeleteQueryRequest\x1a#.FederatedQuery.DeleteQueryResponse\x12Y\n\x0c\x43ontrolQuery\x12#.FederatedQuery.ControlQueryRequest\x1a$.FederatedQuery.ControlQueryResponse\x12\\\n\rGetResultData\x12$.FederatedQuery.GetResultDataRequest\x1a%.FederatedQuery.GetResultDataResponse\x12M\n\x08ListJobs\x12\x1f.FederatedQuery.ListJobsRequest\x1a .FederatedQuery.ListJobsResponse\x12V\n\x0b\x44\x65scribeJob\x12\".FederatedQuery.DescribeJobRequest\x1a#.FederatedQuery.DescribeJobResponse\x12\x65\n\x10\x43reateConnection\x12\'.FederatedQuery.CreateConnectionRequest\x1a(.FederatedQuery.CreateConnectionResponse\x12\x62\n\x0fListConnections\x12&.FederatedQuery.ListConnectionsRequest\x1a\'.FederatedQuery.ListConnectionsResponse\x12k\n\x12\x44\x65scribeConnection\x12).FederatedQuery.DescribeConnectionRequest\x1a*.FederatedQuery.DescribeConnectionResponse\x12\x65\n\x10ModifyConnection\x12\'.FederatedQuery.ModifyConnectionRequest\x1a(.FederatedQuery.ModifyConnectionResponse\x12\x65\n\x10\x44\x65leteConnection\x12\'.FederatedQuery.DeleteConnectionRequest\x1a(.FederatedQuery.DeleteConnectionResponse\x12_\n\x0eTestConnection\x12%.FederatedQuery.TestConnectionRequest\x1a&.FederatedQuery.TestConnectionResponse\x12\\\n\rCreateBinding\x12$.FederatedQuery.CreateBindingRequest\x1a%.FederatedQuery.CreateBindingResponse\x12Y\n\x0cListBindings\x12#.FederatedQuery.ListBindingsRequest\x1a$.FederatedQuery.ListBindingsResponse\x12\x62\n\x0f\x44\x65scribeBinding\x12&.FederatedQuery.DescribeBindingRequest\x1a\'.FederatedQuery.DescribeBindingResponse\x12\\\n\rModifyBinding\x12$.FederatedQuery.ModifyBindingRequest\x1a%.FederatedQuery.ModifyBindingResponse\x12\\\n\rDeleteBinding\x12$.FederatedQuery.DeleteBindingRequest\x1a%.FederatedQuery.DeleteBindingResponseBn\n\'tech.ydb.proto.draft.federated.query.v1ZCgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_FederatedQuery_V1b\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'draft.ydb_federated_query_v1_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\'tech.ydb.proto.draft.federated.query.v1ZCgithub.com/ydb-platform/ydb-go-genproto/draft/Ydb_FederatedQuery_V1'
  _FEDERATEDQUERYSERVICE._serialized_start=98
  _FEDERATEDQUERYSERVICE._serialized_end=2104
# @@protoc_insertion_point(module_scope)



from protos import ydb_operation_pb2 as _ydb_operation_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AddVolatileConfigRequest(_message.Message):
    __slots__ = ["config", "operation_params"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    config: str
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., config: _Optional[str] = ...) -> None: ...

class AddVolatileConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ConfigIdentity(_message.Message):
    __slots__ = ["cluster", "version"]
    CLUSTER_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    cluster: str
    version: int
    def __init__(self, version: _Optional[int] = ..., cluster: _Optional[str] = ...) -> None: ...

class DropConfigRequest(_message.Message):
    __slots__ = ["identity", "operation_params"]
    IDENTITY_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    identity: ConfigIdentity
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., identity: _Optional[_Union[ConfigIdentity, _Mapping]] = ...) -> None: ...

class DropConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetConfigRequest(_message.Message):
    __slots__ = ["operation_params"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class GetConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetConfigResult(_message.Message):
    __slots__ = ["config", "identity", "volatile_configs"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    IDENTITY_FIELD_NUMBER: _ClassVar[int]
    VOLATILE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    config: str
    identity: ConfigIdentity
    volatile_configs: _containers.RepeatedCompositeFieldContainer[VolatileConfig]
    def __init__(self, config: _Optional[str] = ..., volatile_configs: _Optional[_Iterable[_Union[VolatileConfig, _Mapping]]] = ..., identity: _Optional[_Union[ConfigIdentity, _Mapping]] = ...) -> None: ...

class GetMetadataRequest(_message.Message):
    __slots__ = ["operation_params"]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ...) -> None: ...

class GetMetadataResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetMetadataResult(_message.Message):
    __slots__ = ["metadata", "volatile_configs"]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    VOLATILE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    metadata: str
    volatile_configs: _containers.RepeatedCompositeFieldContainer[VolatileConfigMetadata]
    def __init__(self, metadata: _Optional[str] = ..., volatile_configs: _Optional[_Iterable[_Union[VolatileConfigMetadata, _Mapping]]] = ...) -> None: ...

class GetNodeLabelsRequest(_message.Message):
    __slots__ = ["node_id", "operation_params"]
    NODE_ID_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    node_id: int
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., node_id: _Optional[int] = ...) -> None: ...

class GetNodeLabelsResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class GetNodeLabelsResult(_message.Message):
    __slots__ = ["labels"]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    labels: _containers.RepeatedCompositeFieldContainer[YamlLabel]
    def __init__(self, labels: _Optional[_Iterable[_Union[YamlLabel, _Mapping]]] = ...) -> None: ...

class LabelSet(_message.Message):
    __slots__ = ["labels"]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    labels: _containers.RepeatedCompositeFieldContainer[YamlLabelExt]
    def __init__(self, labels: _Optional[_Iterable[_Union[YamlLabelExt, _Mapping]]] = ...) -> None: ...

class RemoveVolatileConfigRequest(_message.Message):
    __slots__ = ["all", "force", "identity", "ids", "operation_params"]
    class IdsToDelete(_message.Message):
        __slots__ = ["ids"]
        IDS_FIELD_NUMBER: _ClassVar[int]
        ids: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, ids: _Optional[_Iterable[int]] = ...) -> None: ...
    ALL_FIELD_NUMBER: _ClassVar[int]
    FORCE_FIELD_NUMBER: _ClassVar[int]
    IDENTITY_FIELD_NUMBER: _ClassVar[int]
    IDS_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    all: bool
    force: bool
    identity: ConfigIdentity
    ids: RemoveVolatileConfigRequest.IdsToDelete
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., identity: _Optional[_Union[ConfigIdentity, _Mapping]] = ..., force: bool = ..., ids: _Optional[_Union[RemoveVolatileConfigRequest.IdsToDelete, _Mapping]] = ..., all: bool = ...) -> None: ...

class RemoveVolatileConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ReplaceConfigRequest(_message.Message):
    __slots__ = ["allow_unknown_fields", "config", "dry_run", "operation_params"]
    ALLOW_UNKNOWN_FIELDS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    DRY_RUN_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    allow_unknown_fields: bool
    config: str
    dry_run: bool
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., config: _Optional[str] = ..., dry_run: bool = ..., allow_unknown_fields: bool = ...) -> None: ...

class ReplaceConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ResolveAllConfigRequest(_message.Message):
    __slots__ = ["config", "operation_params", "verbose_response", "volatile_configs"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    VERBOSE_RESPONSE_FIELD_NUMBER: _ClassVar[int]
    VOLATILE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    config: str
    operation_params: _ydb_operation_pb2.OperationParams
    verbose_response: bool
    volatile_configs: _containers.RepeatedCompositeFieldContainer[VolatileConfig]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., config: _Optional[str] = ..., volatile_configs: _Optional[_Iterable[_Union[VolatileConfig, _Mapping]]] = ..., verbose_response: bool = ...) -> None: ...

class ResolveAllConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ResolveAllConfigResult(_message.Message):
    __slots__ = ["config", "configs"]
    CONFIGS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    config: str
    configs: _containers.RepeatedCompositeFieldContainer[ResolvedConfig]
    def __init__(self, config: _Optional[str] = ..., configs: _Optional[_Iterable[_Union[ResolvedConfig, _Mapping]]] = ...) -> None: ...

class ResolveConfigRequest(_message.Message):
    __slots__ = ["config", "labels", "operation_params", "volatile_configs"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    VOLATILE_CONFIGS_FIELD_NUMBER: _ClassVar[int]
    config: str
    labels: _containers.RepeatedCompositeFieldContainer[YamlLabel]
    operation_params: _ydb_operation_pb2.OperationParams
    volatile_configs: _containers.RepeatedCompositeFieldContainer[VolatileConfig]
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., config: _Optional[str] = ..., volatile_configs: _Optional[_Iterable[_Union[VolatileConfig, _Mapping]]] = ..., labels: _Optional[_Iterable[_Union[YamlLabel, _Mapping]]] = ...) -> None: ...

class ResolveConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class ResolveConfigResult(_message.Message):
    __slots__ = ["config"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    config: str
    def __init__(self, config: _Optional[str] = ...) -> None: ...

class ResolvedConfig(_message.Message):
    __slots__ = ["config", "label_sets"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    LABEL_SETS_FIELD_NUMBER: _ClassVar[int]
    config: str
    label_sets: _containers.RepeatedCompositeFieldContainer[LabelSet]
    def __init__(self, label_sets: _Optional[_Iterable[_Union[LabelSet, _Mapping]]] = ..., config: _Optional[str] = ...) -> None: ...

class SetConfigRequest(_message.Message):
    __slots__ = ["allow_unknown_fields", "config", "dry_run", "operation_params"]
    ALLOW_UNKNOWN_FIELDS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    DRY_RUN_FIELD_NUMBER: _ClassVar[int]
    OPERATION_PARAMS_FIELD_NUMBER: _ClassVar[int]
    allow_unknown_fields: bool
    config: str
    dry_run: bool
    operation_params: _ydb_operation_pb2.OperationParams
    def __init__(self, operation_params: _Optional[_Union[_ydb_operation_pb2.OperationParams, _Mapping]] = ..., config: _Optional[str] = ..., dry_run: bool = ..., allow_unknown_fields: bool = ...) -> None: ...

class SetConfigResponse(_message.Message):
    __slots__ = ["operation"]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    operation: _ydb_operation_pb2.Operation
    def __init__(self, operation: _Optional[_Union[_ydb_operation_pb2.Operation, _Mapping]] = ...) -> None: ...

class VolatileConfig(_message.Message):
    __slots__ = ["config", "id"]
    CONFIG_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    config: str
    id: int
    def __init__(self, id: _Optional[int] = ..., config: _Optional[str] = ...) -> None: ...

class VolatileConfigMetadata(_message.Message):
    __slots__ = ["id", "metadata"]
    ID_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    id: int
    metadata: str
    def __init__(self, id: _Optional[int] = ..., metadata: _Optional[str] = ...) -> None: ...

class YamlLabel(_message.Message):
    __slots__ = ["label", "value"]
    LABEL_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    label: str
    value: str
    def __init__(self, label: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class YamlLabelExt(_message.Message):
    __slots__ = ["label", "type", "value"]
    class LabelType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
    COMMON: YamlLabelExt.LabelType
    EMPTY: YamlLabelExt.LabelType
    LABEL_FIELD_NUMBER: _ClassVar[int]
    NOT_SET: YamlLabelExt.LabelType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    UNSPECIFIED: YamlLabelExt.LabelType
    VALUE_FIELD_NUMBER: _ClassVar[int]
    label: str
    type: YamlLabelExt.LabelType
    value: str
    def __init__(self, label: _Optional[str] = ..., type: _Optional[_Union[YamlLabelExt.LabelType, str]] = ..., value: _Optional[str] = ...) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v5.protos import ydb_operation_pb2 as protos_dot_ydb__operation__pb2


class OperationServiceStub(object):
    """All rpc calls to YDB are allowed to be asynchronous. Response message
    of an rpc call contains Operation structure and OperationService
    is used for polling operation completion.

    Operation has a field 'ready' to notify client if operation has been
    completed or not. If result is ready a client has to handle 'result' field,
    otherwise it is expected that client continues polling result via
    GetOperation rpc of OperationService. Polling is made via unique
    operation id provided in 'id' field of Operation.

    Note: Currently some operations have synchronous implementation and their result
    is available when response is obtained. But a client must not make any
    assumptions about synchronous or asynchronous nature of any operation and
    be ready to poll operation status.

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetOperation = channel.unary_unary(
                '/Ydb.Operation.V1.OperationService/GetOperation',
                request_serializer=protos_dot_ydb__operation__pb2.GetOperationRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__operation__pb2.GetOperationResponse.FromString,
                )
        self.CancelOperation = channel.unary_unary(
                '/Ydb.Operation.V1.OperationService/CancelOperation',
                request_serializer=protos_dot_ydb__operation__pb2.CancelOperationRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__operation__pb2.CancelOperationResponse.FromString,
                )
        self.ForgetOperation = channel.unary_unary(
                '/Ydb.Operation.V1.OperationService/ForgetOperation',
                request_serializer=protos_dot_ydb__operation__pb2.ForgetOperationRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__operation__pb2.ForgetOperationResponse.FromString,
                )
        self.ListOperations = channel.unary_unary(
                '/Ydb.Operation.V1.OperationService/ListOperations',
                request_serializer=protos_dot_ydb__operation__pb2.ListOperationsRequest.SerializeToString,
                response_deserializer=protos_dot_ydb__operation__pb2.ListOperationsResponse.FromString,
                )


class OperationServiceServicer(object):
    """All rpc calls to YDB are allowed to be asynchronous. Response message
    of an rpc call contains Operation structure and OperationService
    is used for polling operation completion.

    Operation has a field 'ready' to notify client if operation has been
    completed or not. If result is ready a client has to handle 'result' field,
    otherwise it is expected that client continues polling result via
    GetOperation rpc of OperationService. Polling is made via unique
    operation id provided in 'id' field of Operation.

    Note: Currently some operations have synchronous implementation and their result
    is available when response is obtained. But a client must not make any
    assumptions about synchronous or asynchronous nature of any operation and
    be ready to poll operation status.

    """

    def GetOperation(self, request, context):
        """Check status for a given operation.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelOperation(self, request, context):
        """Starts cancellation of a long-running operation,
        Clients can use GetOperation to check whether the cancellation succeeded
        or whether the operation completed despite cancellation.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ForgetOperation(self, request, context):
        """Forgets long-running operation. It does not cancel the operation and returns
        an error if operation was not completed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListOperations(self, request, context):
        """Lists operations that match the specified filter in the request.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_OperationServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetOperation': grpc.unary_unary_rpc_method_handler(
                    servicer.GetOperation,
                    request_deserializer=protos_dot_ydb__operation__pb2.GetOperationRequest.FromString,
                    response_serializer=protos_dot_ydb__operation__pb2.GetOperationResponse.SerializeToString,
            ),
            'CancelOperation': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelOperation,
                    request_deserializer=protos_dot_ydb__operation__pb2.CancelOperationRequest.FromString,
                    response_serializer=protos_dot_ydb__operation__pb2.CancelOperationResponse.SerializeToString,
            ),
            'ForgetOperation': grpc.unary_unary_rpc_method_handler(
                    servicer.ForgetOperation,
                    request_deserializer=protos_dot_ydb__operation__pb2.ForgetOperationRequest.FromString,
                    response_serializer=protos_dot_ydb__operation__pb2.ForgetOperationResponse.SerializeToString,
            ),
            'ListOperations': grpc.unary_unary_rpc_method_handler(
                    servicer.ListOperations,
                    request_deserializer=protos_dot_ydb__operation__pb2.ListOperationsRequest.FromString,
                    response_serializer=protos_dot_ydb__operation__pb2.ListOperationsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.Operation.V1.OperationService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class OperationService(object):
    """All rpc calls to YDB are allowed to be asynchronous. Response message
    of an rpc call contains Operation structure and OperationService
    is used for polling operation completion.

    Operation has a field 'ready' to notify client if operation has been
    completed or not. If result is ready a client has to handle 'result' field,
    otherwise it is expected that client continues polling result via
    GetOperation rpc of OperationService. Polling is made via unique
    operation id provided in 'id' field of Operation.

    Note: Currently some operations have synchronous implementation and their result
    is available when response is obtained. But a client must not make any
    assumptions about synchronous or asynchronous nature of any operation and
    be ready to poll operation status.

    """

    @staticmethod
    def GetOperation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Operation.V1.OperationService/GetOperation',
            protos_dot_ydb__operation__pb2.GetOperationRequest.SerializeToString,
            protos_dot_ydb__operation__pb2.GetOperationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CancelOperation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Operation.V1.OperationService/CancelOperation',
            protos_dot_ydb__operation__pb2.CancelOperationRequest.SerializeToString,
            protos_dot_ydb__operation__pb2.CancelOperationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ForgetOperation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Operation.V1.OperationService/ForgetOperation',
            protos_dot_ydb__operation__pb2.ForgetOperationRequest.SerializeToString,
            protos_dot_ydb__operation__pb2.ForgetOperationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListOperations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.Operation.V1.OperationService/ListOperations',
            protos_dot_ydb__operation__pb2.ListOperationsRequest.SerializeToString,
            protos_dot_ydb__operation__pb2.ListOperationsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# flake8: noqa
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ydb_scheme_v1.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ydb._grpc.v3.protos import ydb_scheme_pb2 as protos_dot_ydb__scheme__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ydb_scheme_v1.proto',
  package='Ydb.Scheme.V1',
  syntax='proto3',
  serialized_options=b'\n\030tech.ydb.proto.scheme.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Scheme_V1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x13ydb_scheme_v1.proto\x12\rYdb.Scheme.V1\x1a\x17protos/ydb_scheme.proto2\xcc\x03\n\rSchemeService\x12T\n\rMakeDirectory\x12 .Ydb.Scheme.MakeDirectoryRequest\x1a!.Ydb.Scheme.MakeDirectoryResponse\x12Z\n\x0fRemoveDirectory\x12\".Ydb.Scheme.RemoveDirectoryRequest\x1a#.Ydb.Scheme.RemoveDirectoryResponse\x12T\n\rListDirectory\x12 .Ydb.Scheme.ListDirectoryRequest\x1a!.Ydb.Scheme.ListDirectoryResponse\x12Q\n\x0c\x44\x65scribePath\x12\x1f.Ydb.Scheme.DescribePathRequest\x1a .Ydb.Scheme.DescribePathResponse\x12`\n\x11ModifyPermissions\x12$.Ydb.Scheme.ModifyPermissionsRequest\x1a%.Ydb.Scheme.ModifyPermissionsResponseBQ\n\x18tech.ydb.proto.scheme.v1Z5github.com/ydb-platform/ydb-go-genproto/Ydb_Scheme_V1b\x06proto3'
  ,
  dependencies=[protos_dot_ydb__scheme__pb2.DESCRIPTOR,])



_sym_db.RegisterFileDescriptor(DESCRIPTOR)


DESCRIPTOR._options = None

_SCHEMESERVICE = _descriptor.ServiceDescriptor(
  name='SchemeService',
  full_name='Ydb.Scheme.V1.SchemeService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=64,
  serialized_end=524,
  methods=[
  _descriptor.MethodDescriptor(
    name='MakeDirectory',
    full_name='Ydb.Scheme.V1.SchemeService.MakeDirectory',
    index=0,
    containing_service=None,
    input_type=protos_dot_ydb__scheme__pb2._MAKEDIRECTORYREQUEST,
    output_type=protos_dot_ydb__scheme__pb2._MAKEDIRECTORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RemoveDirectory',
    full_name='Ydb.Scheme.V1.SchemeService.RemoveDirectory',
    index=1,
    containing_service=None,
    input_type=protos_dot_ydb__scheme__pb2._REMOVEDIRECTORYREQUEST,
    output_type=protos_dot_ydb__scheme__pb2._REMOVEDIRECTORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListDirectory',
    full_name='Ydb.Scheme.V1.SchemeService.ListDirectory',
    index=2,
    containing_service=None,
    input_type=protos_dot_ydb__scheme__pb2._LISTDIRECTORYREQUEST,
    output_type=protos_dot_ydb__scheme__pb2._LISTDIRECTORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DescribePath',
    full_name='Ydb.Scheme.V1.SchemeService.DescribePath',
    index=3,
    containing_service=None,
    input_type=protos_dot_ydb__scheme__pb2._DESCRIBEPATHREQUEST,
    output_type=protos_dot_ydb__scheme__pb2._DESCRIBEPATHRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ModifyPermissions',
    full_name='Ydb.Scheme.V1.SchemeService.ModifyPermissions',
    index=4,
    containing_service=None,
    input_type=protos_dot_ydb__scheme__pb2._MODIFYPERMISSIONSREQUEST,
    output_type=protos_dot_ydb__scheme__pb2._MODIFYPERMISSIONSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_SCHEMESERVICE)

DESCRIPTOR.services_by_name['SchemeService'] = _SCHEMESERVICE

# @@protoc_insertion_point(module_scope)



import asyncio
import threading

import pytest

import ydb
from ydb.aio.coordination import CoordinationClient as AioCoordinationClient
from ydb import StatusCode

from ydb.coordination import (
    NodeConfig,
    ConsistencyMode,
    RateLimiterCountersMode,
    CoordinationClient,
)


@pytest.fixture
def sync_coordination_node(driver_sync):
    client = CoordinationClient(driver_sync)
    node_path = "/local/test_node"

    try:
        client.delete_node(node_path)
    except ydb.SchemeError:
        pass

    config = NodeConfig(
        session_grace_period_millis=1000,
        attach_consistency_mode=ConsistencyMode.STRICT,
        read_consistency_mode=ConsistencyMode.STRICT,
        rate_limiter_counters_mode=RateLimiterCountersMode.UNSET,
        self_check_period_millis=0,
    )
    client.create_node(node_path, config)

    yield client, node_path, config

    try:
        client.delete_node(node_path)
    except ydb.SchemeError:
        pass


@pytest.fixture
async def async_coordination_node(aio_connection):
    client = AioCoordinationClient(aio_connection)
    node_path = "/local/test_node"

    try:
        await client.delete_node(node_path)
    except ydb.SchemeError:
        pass

    config = NodeConfig(
        session_grace_period_millis=1000,
        attach_consistency_mode=ConsistencyMode.STRICT,
        read_consistency_mode=ConsistencyMode.STRICT,
        rate_limiter_counters_mode=RateLimiterCountersMode.UNSET,
        self_check_period_millis=0,
    )
    await client.create_node(node_path, config)

    yield client, node_path, config

    try:
        await client.delete_node(node_path)
    except ydb.SchemeError:
        pass


class TestCoordination:
    def test_coordination_node_lifecycle(self, sync_coordination_node):
        client, node_path, initial_config = sync_coordination_node

        node_conf = client.describe_node(node_path)
        assert node_conf == initial_config

        updated_config = NodeConfig(
            session_grace_period_millis=12345,
            attach_consistency_mode=ConsistencyMode.STRICT,
            read_consistency_mode=ConsistencyMode.RELAXED,
            rate_limiter_counters_mode=RateLimiterCountersMode.DETAILED,
            self_check_period_millis=10,
        )
        client.alter_node(node_path, updated_config)

        node_conf = client.describe_node(node_path)
        assert node_conf == updated_config

        client.delete_node(node_path)

        with pytest.raises(ydb.SchemeError):
            client.describe_node(node_path)

    async def test_coordination_node_lifecycle_async(self, async_coordination_node):
        client, node_path, initial_config = async_coordination_node

        node_conf = await client.describe_node(node_path)
        assert node_conf == initial_config

        updated_config = NodeConfig(
            session_grace_period_millis=12345,
            attach_consistency_mode=ConsistencyMode.STRICT,
            read_consistency_mode=ConsistencyMode.RELAXED,
            rate_limiter_counters_mode=RateLimiterCountersMode.DETAILED,
            self_check_period_millis=10,
        )
        await client.alter_node(node_path, updated_config)

        node_conf = await client.describe_node(node_path)
        assert node_conf == updated_config

        await client.delete_node(node_path)

        with pytest.raises(ydb.SchemeError):
            await client.describe_node(node_path)

    async def test_coordination_lock_describe_full_async(self, async_coordination_node):
        client, node_path, _ = async_coordination_node

        async with client.session(node_path) as node:
            lock = node.semaphore("test_lock")

            desc = await lock.describe()
            assert desc.status == StatusCode.NOT_FOUND

            async with lock:
                pass

            desc = await lock.describe()
            assert desc.data == b""

            await lock.update(new_data=b"world")

            desc2 = await lock.describe()
            assert desc2.data == b"world"

    def test_coordination_lock_describe_full(self, sync_coordination_node):
        client, node_path, _ = sync_coordination_node

        with client.session(node_path) as node:
            lock = node.semaphore("test_lock")

            desc = lock.describe()
            assert desc.status == StatusCode.NOT_FOUND

            with lock:
                pass

            desc = lock.describe()
            assert desc.data == b""

            lock.update(new_data=b"world")

            desc2 = lock.describe()
            assert desc2.data == b"world"

    async def test_coordination_lock_racing_async(self, async_coordination_node):
        client, node_path, _ = async_coordination_node
        timeout = 5

        async with client.session(node_path) as node:
            lock2_started = asyncio.Event()
            lock2_acquired = asyncio.Event()
            lock2_release = asyncio.Event()

            async def second_lock_task():
                lock2_started.set()
                async with node.semaphore("test_lock"):
                    lock2_acquired.set()
                    await lock2_release.wait()

            async with node.semaphore("test_lock"):
                t2 = asyncio.create_task(second_lock_task())
                await asyncio.wait_for(lock2_started.wait(), timeout=timeout)

            await asyncio.wait_for(lock2_acquired.wait(), timeout=timeout)
            lock2_release.set()
            await asyncio.wait_for(t2, timeout=timeout)

    def test_coordination_lock_racing(self, sync_coordination_node):
        client, node_path, _ = sync_coordination_node
        timeout = 5

        with client.session(node_path) as node:
            lock2_started = threading.Event()
            lock2_acquired = threading.Event()
            lock2_release = threading.Event()

            def second_lock_task():
                lock2_started.set()
                with node.semaphore("test_lock"):
                    lock2_acquired.set()
                    lock2_release.wait(timeout)

            with node.semaphore("test_lock"):
                t2 = threading.Thread(target=second_lock_task)
                t2.start()

                assert lock2_started.wait(timeout)

            assert lock2_acquired.wait(timeout)
            lock2_release.set()
            t2.join(timeout)

    async def test_coordination_reconnect_async(self, async_coordination_node):
        client, node_path, _ = async_coordination_node

        async with client.session(node_path) as node:
            lock = node.semaphore("test_lock")

            async with lock:
                pass

            await node._reconnector._stream.close()

            async with lock:
                pass

    async def test_same_lock_cannot_be_acquired_twice(self, async_coordination_node):
        client, node_path, _ = async_coordination_node

        async with client.session(node_path) as node:
            lock1 = node.semaphore("lock1")
            lock1_1 = node.semaphore("lock1")

            await lock1.acquire()

            acquire_task = asyncio.create_task(lock1_1.acquire())

            assert not acquire_task.done()

            await lock1.release()
            await asyncio.wait_for(acquire_task, timeout=5)



import concurrent.futures
import datetime
import enum
import itertools
import uuid
from dataclasses import dataclass
from enum import Enum
from typing import List, Union, Optional, Any, Dict

import typing

import ydb.aio
from .._grpc.grpcwrapper.ydb_topic import StreamWriteMessage
from .._grpc.grpcwrapper.ydb_topic import TransactionIdentity
from .._grpc.grpcwrapper.common_utils import IToProto
from .._grpc.grpcwrapper.ydb_topic_public_types import PublicCodec
from .. import connection

Message = typing.Union["PublicMessage", "PublicMessage.SimpleSourceType"]


@dataclass
class PublicWriterSettings:
    """
    Settings for topic writer.

    order of fields IS NOT stable, use keywords only
    """

    topic: str
    producer_id: Optional[str] = None
    session_metadata: Optional[Dict[str, str]] = None
    partition_id: Optional[int] = None
    auto_seqno: bool = True
    auto_created_at: bool = True
    codec: Optional[PublicCodec] = None  # default mean auto-select
    encoder_executor: Optional[concurrent.futures.Executor] = None  # default shared client executor pool
    encoders: Optional[typing.Mapping[PublicCodec, typing.Callable[[bytes], bytes]]] = None
    update_token_interval: Union[int, float] = 3600

    def __post_init__(self):
        if self.producer_id is None:
            self.producer_id = uuid.uuid4().hex


@dataclass
class PublicWriteResult:
    @dataclass(eq=True)
    class Written:
        __slots__ = "offset"
        offset: int

    @dataclass(eq=True)
    class Skipped:
        pass

    @dataclass(eq=True)
    class WrittenInTx:
        pass


PublicWriteResultTypes = Union[PublicWriteResult.Written, PublicWriteResult.Skipped, PublicWriteResult.WrittenInTx]


class WriterSettings(PublicWriterSettings):
    def __init__(self, settings: PublicWriterSettings):
        self.__dict__ = settings.__dict__.copy()

    def create_init_request(self) -> StreamWriteMessage.InitRequest:
        return StreamWriteMessage.InitRequest(
            path=self.topic,
            producer_id=self.producer_id,
            write_session_meta=self.session_metadata,
            partitioning=self.get_partitioning(),
            get_last_seq_no=True,
        )

    def get_partitioning(self) -> StreamWriteMessage.PartitioningType:
        if self.partition_id is not None:
            return StreamWriteMessage.PartitioningPartitionID(self.partition_id)
        return StreamWriteMessage.PartitioningMessageGroupID(self.producer_id)


class SendMode(Enum):
    ASYNC = 1
    SYNC = 2


@dataclass
class PublicWriterInitInfo:
    __slots__ = ("last_seqno", "supported_codecs")
    last_seqno: Optional[int]
    supported_codecs: List[PublicCodec]


class PublicMessage:
    seqno: Optional[int]
    created_at: Optional[datetime.datetime]
    data: "PublicMessage.SimpleSourceType"
    metadata_items: Optional[Dict[str, "PublicMessage.SimpleSourceType"]]

    SimpleSourceType = Union[str, bytes]  # Will be extend

    def __init__(
        self,
        data: SimpleSourceType,
        *,
        metadata_items: Optional[Dict[str, "PublicMessage.SimpleSourceType"]] = None,
        seqno: Optional[int] = None,
        created_at: Optional[datetime.datetime] = None,
    ):
        self.seqno = seqno
        self.created_at = created_at
        self.data = data
        self.metadata_items = metadata_items

    @staticmethod
    def _create_message(data: Message) -> "PublicMessage":
        if isinstance(data, PublicMessage):
            return data
        return PublicMessage(data=data)


class InternalMessage(StreamWriteMessage.WriteRequest.MessageData, IToProto):
    codec: PublicCodec

    def __init__(self, mess: PublicMessage):
        metadata_items = mess.metadata_items or {}
        super().__init__(
            seq_no=mess.seqno,
            created_at=mess.created_at,
            data=mess.data,
            metadata_items=metadata_items,
            uncompressed_size=len(mess.data),
            partitioning=None,
        )
        self.codec = PublicCodec.RAW

    def _get_bytes(self, obj: Optional[PublicMessage.SimpleSourceType]) -> bytes:
        if obj is None:
            return bytes()
        if isinstance(obj, bytes):
            return obj
        if isinstance(obj, str):
            return obj.encode("utf-8")
        raise ValueError("Bad data type")

    def get_data_bytes(self) -> bytes:
        return self._get_bytes(self.data)

    def to_message_data(self) -> StreamWriteMessage.WriteRequest.MessageData:
        data = self.get_data_bytes()
        metadata_items = {key: self._get_bytes(value) for key, value in self.metadata_items.items()}
        return StreamWriteMessage.WriteRequest.MessageData(
            seq_no=self.seq_no,
            created_at=self.created_at,
            data=data,
            metadata_items=metadata_items,
            uncompressed_size=len(data),
            partitioning=None,  # unsupported by server now
        )


class MessageSendResult:
    offset: Optional[int]
    write_status: "MessageWriteStatus"


class MessageWriteStatus(enum.Enum):
    Written = 1
    AlreadyWritten = 2


class RetryPolicy:
    connection_timeout_sec: float
    overload_timeout_sec: float
    retry_access_denied: bool = False


class TopicWriterError(ydb.Error):
    def __init__(self, message: str):
        super(TopicWriterError, self).__init__(message)


class TopicWriterClosedError(ydb.Error):
    def __init__(self):
        super().__init__("Topic writer already closed")


class TopicWriterRepeatableError(TopicWriterError):
    pass


class TopicWriterStopped(TopicWriterError):
    def __init__(self):
        super(TopicWriterStopped, self).__init__("topic writer was stopped by call close")


def default_serializer_message_content(data: Any) -> bytes:
    if data is None:
        return bytes()
    if isinstance(data, bytes):
        return data
    if isinstance(data, bytearray):
        return bytes(data)
    if isinstance(data, str):
        return data.encode(encoding="utf-8")
    raise ValueError("can't serialize type %s to bytes" % type(data))


def messages_to_proto_requests(
    messages: List[InternalMessage],
    tx_identity: Optional[TransactionIdentity],
) -> List[StreamWriteMessage.FromClient]:

    groups = _split_messages_for_send(messages)

    res = []  # type: List[StreamWriteMessage.FromClient]
    for group in groups:
        req = StreamWriteMessage.FromClient(
            StreamWriteMessage.WriteRequest(
                messages=list(map(InternalMessage.to_message_data, group)),
                codec=group[0].codec,
                tx_identity=tx_identity,
            )
        )
        res.append(req)
    return res


_max_int = 2**63 - 1

_message_data_overhead = (
    StreamWriteMessage.FromClient(
        StreamWriteMessage.WriteRequest(
            messages=[
                StreamWriteMessage.WriteRequest.MessageData(
                    seq_no=_max_int,
                    created_at=datetime.datetime(3000, 1, 1, 1, 1, 1, 1),
                    data=bytes(1),
                    metadata_items={},
                    uncompressed_size=_max_int,
                    partitioning=StreamWriteMessage.PartitioningMessageGroupID(
                        message_group_id="a" * 100,
                    ),
                ),
            ],
            codec=20000,
            tx_identity=None,
        )
    )
    .to_proto()
    .ByteSize()
)


def _split_messages_for_send(
    messages: List[InternalMessage],
) -> List[List[InternalMessage]]:
    codec_groups = []  # type: List[List[InternalMessage]]
    for _, messages in itertools.groupby(messages, lambda x: x.codec):
        codec_groups.append(list(messages))

    res = []  # type: List[List[InternalMessage]]
    for codec_group in codec_groups:
        group_by_size = _split_messages_by_size_with_default_overhead(codec_group)
        res.extend(group_by_size)
    return res


def _split_messages_by_size_with_default_overhead(
    messages: List[InternalMessage],
) -> List[List[InternalMessage]]:
    def get_message_size(msg: InternalMessage):
        return len(msg.data) + _message_data_overhead

    return _split_messages_by_size(messages, connection._DEFAULT_MAX_GRPC_MESSAGE_SIZE, get_message_size)


def _split_messages_by_size(
    messages: List[InternalMessage],
    split_size: int,
    get_msg_size: typing.Callable[[InternalMessage], int],
) -> List[List[InternalMessage]]:
    res = []
    group = []
    group_size = 0

    for msg in messages:
        msg_size = get_msg_size(msg)

        if len(group) == 0:
            group.append(msg)
            group_size += msg_size
        elif group_size + msg_size <= split_size:
            group.append(msg)
            group_size += msg_size
        else:
            res.append(group)
            group = [msg]
            group_size = msg_size

    if len(group) > 0:
        res.append(group)

    return res



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ydb._grpc.v3.draft.protos import ydb_object_storage_pb2 as draft_dot_protos_dot_ydb__object__storage__pb2


class ObjectStorageServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.List = channel.unary_unary(
                '/Ydb.ObjectStorage.V1.ObjectStorageService/List',
                request_serializer=draft_dot_protos_dot_ydb__object__storage__pb2.ListingRequest.SerializeToString,
                response_deserializer=draft_dot_protos_dot_ydb__object__storage__pb2.ListingResponse.FromString,
                )


class ObjectStorageServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def List(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ObjectStorageServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=draft_dot_protos_dot_ydb__object__storage__pb2.ListingRequest.FromString,
                    response_serializer=draft_dot_protos_dot_ydb__object__storage__pb2.ListingResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Ydb.ObjectStorage.V1.ObjectStorageService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ObjectStorageService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/Ydb.ObjectStorage.V1.ObjectStorageService/List',
            draft_dot_protos_dot_ydb__object__storage__pb2.ListingRequest.SerializeToString,
            draft_dot_protos_dot_ydb__object__storage__pb2.ListingResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
